from decimal import Decimal
from ofxparse.ofxparse import Transaction, InvestmentTransaction

def clean_ofx_id(ofxid):
    ofxid = ofxid.replace('/', '_')
    ofxid = ofxid.replace('$', '_')
    ofxid = ofxid.replace(' ', '_')
    return ofxid

class Formatter(object):
    def __init__(self, account, name, ledger=None):
        self.acctid=account.account_id
        self.currency=account.statement.currency
        self.fid=account.institution.fid
        self.name = name
        self.ledger = ledger

    def mk_dynamic_account(self, txn):
        if self.ledger is None:
            return "Expenses:Misc"
        else:
            payee = self.format_payee(txn)
            account = self.ledger.get_account_by_payee(payee)
            if account is None:
                return "Expenses:Misc"
            else:
                return account

    def format_amount(self, amount, reverse=False, unlimited=False):
        currency = self.currency.upper()
        if currency == "USD": currency = "$"
        if unlimited:
            amt = str(abs(amount))
        else:
            amt = "%0.2f"%(abs(amount))
        if amount.is_signed() != reverse:
            return "-%s%s"%(currency, amt)
        else:
            return "%s%s"%(currency, amt)

    def format_payee(self, txn):
        if (txn.payee is None) or txn.memo.startswith(txn.payee):
            return txn.memo
        elif (txn.memo is None) or txn.payee.startswith(txn.memo):
            return txn.payee
        else:
            return "%s %s"%(txn.payee, txn.memo)

    def format_txn_line(self, acct, amt, extra=None):
        space_count = 50 - len(acct) - len(amt)
        if extra:
            extra_str = " %s"%(extra)
        else:
            extra_str = ""
        return "  %s%s%s%s\n"%(acct, (" " * space_count), amt, extra_str)

    def format_txn(self, txn):
        retval = ""
        ofxid = clean_ofx_id("%s.%s.%s"%(self.fid, self.acctid, txn.id))
        if isinstance(txn, Transaction):
            date = "%s"%(txn.date.strftime("%Y/%m/%d"))
            retval += "%s %s\n"%(date, self.format_payee(txn))
            retval += "  ; ofxid: %s\n"%(ofxid)
            retval += self.format_txn_line(self.name, self.format_amount(txn.amount))
            retval += self.format_txn_line(self.mk_dynamic_account(txn), self.format_amount(txn.amount, reverse=True))
        elif isinstance(txn, InvestmentTransaction):
            trade_date = "%s"%(txn.tradeDate.strftime("%Y/%m/%d"))
            if txn.settleDate is not None:
                retval = "%s=%s %s\n"%(txn.tradeDate.strftime("%Y/%m/%d"), txn.settleDate.strftime("%Y/%m/%d"), txn.memo)
            else:
                retval = "%s %s\n"%(txn.tradeDate.strftime("%Y/%m/%d"), txn.memo)
            retval += "  ; ofxid: %s\n"%(ofxid)
            retval += self.format_txn_line(self.name, str(txn.units), 
                                           "%s @ %s"%(txn.security, self.format_amount(txn.unit_price, unlimited=True)))
            retval += self.format_txn_line(self.name, self.format_amount(txn.units * txn.unit_price, reverse=True))
        return retval



import xml.etree.ElementTree as ET
import sys
import time
import subprocess
from subprocess import Popen, PIPE
from threading import Thread
from Queue import Queue, Empty
from ledgerautosync.formatter import clean_ofx_id
import logging

def clean_payee(payee):
    payee = payee.replace('/', '\\/')
    payee = payee.replace('%', '')
    return payee

def enqueue_output(out, queue):
    item = ""
    buff = ""
    while (buff != None):
        buff = out.read(1)
        if (buff != None):
            item += buff
        if item.endswith("] "): # prompt
            queue.put(item[0:-2])
            item = ""
    out.close()

def mk_ledger(ledger_file=None):
    if subprocess.call("which ledger > /dev/null", shell=True) == 0:
        return Ledger(ledger_file)
    elif subprocess.call("which hledger > /dev/null", shell=True) == 0:
        return HLedger(ledger_file)
    else:
        raise Exception("Neither ledger nor hledger found!")

class Ledger(object):
    def __init__(self, ledger_file=None):
        on_posix = 'posix' in sys.builtin_module_names
        args = ["ledger"]
        if ledger_file is not None:
            args += ["-f", ledger_file]
        self.p = Popen(args, bufsize=1, stdin=PIPE, stdout=PIPE,
                       close_fds=on_posix)
        self.q = Queue()
        self.t = Thread(target=enqueue_output, args=(self.p.stdout, self.q))
        self.t.daemon = True # thread dies with the program
        self.t.start()
        # read output until prompt
        self.q.get()

    def run(self, cmd):
        self.p.stdin.write("xml %s\n"%(cmd))
        return ET.fromstring(self.q.get())

    def get_transaction(self, q):
        d = self.run("reg %s"%(q)).findall('.//transactions/transaction')
        if len(d) == 0:
            return None
        else:
            return d[0]

    def check_transaction_by_ofxid(self, ofxid):
        return (self.get_transaction("meta ofxid='%s'"%(clean_ofx_id(ofxid))) != None)
    
    def get_account_by_payee(self, payee):
        txn = self.get_transaction("payee '%s'"%(clean_payee(payee)))
        if txn is None: return None
        else: return txn.findall('./postings/posting[2]/account/name')[0].text

class HLedger(object):
    def __init__(self, ledger_file=None):
        self.args = ["hledger"]
        if ledger_file is not None:
            self.args += ["-f", ledger_file]
        
    def check_transaction_by_ofxid(self, ofxid):
        return (subprocess.check_output(self.args + ["reg", "tag:ofxid=%s"%(ofxid)]) != '')

    def get_account_by_payee(self, payee):
        lines = subprocess.check_output(self.args + ["bal", "desc:'%s'"%(payee), "--format", "%(account)"]).splitlines()
        if (len(lines) < 4): return None
        else: return lines[1]



from ofxparse import OfxParser
from StringIO import StringIO
import logging

class Synchronizer(object):
    def __init__(self, ledger):
        self.ledger = ledger

    def parse_file(self, path, accountname=None):
        ofx = OfxParser.parse(file(path))
        return (ofx, self.filter(ofx))

    def is_txn_synced(self, acctid, txn):
        ofxid = "%s.%s"%(acctid, txn.id)
        return self.ledger.check_transaction_by_ofxid(ofxid)
    
    def filter(self, ofx):
        txns = ofx.account.statement.transactions
        acctid = ofx.account.account_id
        return [ txn for txn in txns if not(self.is_txn_synced(acctid, txn)) ]

    def get_new_txns(self, acct, max_days=999999, resync=False):
        if resync or (max_days < 7):
            days = max_days
        else:
            days = 7
        last_txns_len = 0
        while (True):
            logging.debug("Downloading %d days of transactions (max_days=%d)."%(days, max_days))
            raw = acct.download(days)
            ofx = OfxParser.parse(raw)
            txns = ofx.account.statement.transactions
            new_txns = self.filter(ofx)
            if ((len(txns) > 0) and (last_txns_len == len(txns))):
                # not getting more txns than last time; we have
                # reached the beginning
                logging.debug("Not getting more txns than last time, done.")
                return (ofx, new_txns)
            elif (len(txns) > len(new_txns)) or (days >= max_days):
                # got more txns than were new or hit max_days, we've
                # reached a stopping point
                if (days >= max_days):
                    logging.debug("Hit max days.")
                else:
                    logging.debug("Got some stale txns.")
                return (ofx, new_txns)
            else:
                # all txns were new, increase how far back we go
                days = days * 2
                if (days > max_days): days = max_days
                logging.debug("Increasing days ago to %d."%(days))
                last_txns_len = len(txns)



import datetime
import time
from ofxclient.config import OfxConfig
from ofxparse import OfxParser
import argparse
from ofxclient.client import Client
from formatter import Formatter
from ledgerautosync.sync import Synchronizer
from ledgerautosync.ledger import mk_ledger
import logging
import sys

def sync(ledger, config, max_days=90, resync=False):
    sync = Synchronizer(ledger)
    for acct in config.accounts():
        (ofx, txns) = sync.get_new_txns(acct, resync=resync, max_days=max_days)
        formatter = Formatter(account=ofx.account, name=acct.description)
        for txn in txns:
            print formatter.format_txn(txn)

def import_ofx(ledger, path, accountname=None):
    sync = Synchronizer(ledger)
    (ofx, txns) = sync.parse_file(path)
    if accountname is None:
        accountname = "%s:%s"%(ofx.account.institution.organization, ofx.account.account_id)
    formatter = Formatter(account=ofx.account, name=accountname)
    for txn in txns:
        print formatter.format_txn(txn)
        
def run(args=None):
    if args is None: args = sys.argv[1:]

    parser = argparse.ArgumentParser(description='Synchronize ledger.')
    parser.add_argument('-m', '--max', type=int, default=90,
                        help='maximum number of days to process')
    parser.add_argument('-r', '--resync', action='store_true', default=False,
                        help='do not stop until max days reached')
    parser.add_argument('PATH', nargs='?', help='do not sync; import from OFX file')
    parser.add_argument('-a', '--account', type=str, default=None,
                        help='set account name for import')
    parser.add_argument('-l', '--ledger', type=str, default=None,
                        help='specify ledger file to READ for syncing')
    parser.add_argument('-d', '--debug', action='store_true', default=False,
                        help='enable debug logging')
    args = parser.parse_args(args)
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    ledger = mk_ledger(args.ledger)
    if args.PATH is None:
        config = OfxConfig()
        sync(ledger, config, max_days=args.max, resync=args.resync)
    else:
        import_ofx(ledger, args.PATH, args.account)

if __name__ == '__main__':
    run()



from setuptools import setup, find_packages

with open('README.rst') as file:
    long_description = file.read()

setup(name='ledger-autosync',
      version="0.1.0",
      description="Automatically sync your bank's data with ledger",
      long_description=long_description,
      classifiers=[
          "Operating System :: OS Independent",
          "Programming Language :: Python",
      ],
      author='Erik Hetzner',
      author_email='egh@e6h.org',
      url='https://bitbucket.org/egh/ledger-autosync',
      packages=find_packages(exclude=[
          'tests']),
      entry_points={
          'console_scripts': [
              'ledger-autosync = ledgerautosync.cli:run'
          ]
      },
      install_requires=[
          "ofxclient",
      ],
      test_suite='tests',
      )
